/************************************************************************/
/*  Routines to select best input stream.				*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: find_best_stream.c,v 1.2 2000/10/26 13:39:23 doug Exp $ ";
#endif

#include <stdio.h>
#include <math.h>

#include "qlib2.h"

#include "params.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"

/************************************************************************/
/*  find_best_stream:							*/
/*	Find the best input stream from which to read.			*/
/************************************************************************/
ST_INFO *find_best_stream 
   (ST_INFO	*in_p,		/* ptr to current input stream.		*/
    ST_INFO	*out_p,		/* ptr to output stream.		*/
    INT_TIME	*stime,		/* ptr to desired start time, or NULL.	*/
    INT_TIME	*etime,		/* ptr to desired end time, or NULL.	*/
    int		bias)		/* bias flag for best stream selection.*/
{
    ST_INFO	*st_p;
    ST_INFO	*best_p = NULL;
    double	cmp, tdif;
    double	best_diffnum, diffnum;
    int		usecs_per_point;
    int		seconds, usecs;
    INT_TIME	exptime;
    int b2000_recnum;

    b2000_recnum = (out_p->cur_hdr) ? out_p->cur_hdr->xm2 : IGNORE2000;

    /*	If either we have a start time OR we have no current stream,	*/
    /*	then position all input streams.				*/
    if (stime != NULL || in_p == NULL) 
	for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
	    /*	If possible, use header time for a fast check to	*/
	    /*	determine whether desired time interval is in this file.*/
	    if ( (! ignore_hdr) && (! missing_time(st_p->begtime)) && 
		(! missing_time(st_p->endtime)) && 
		(stime != NULL | etime != NULL) ) {
		if ((stime != NULL && tdiff(st_p->endtime,*stime) < 0) ||
		    (etime != NULL && tdiff(st_p->begtime,*etime) > 0)) {
		    st_p = close_stream(st_p);
		    continue;
		}
	    }
	    if (position_stream (st_p, stime, etime, IGNORE2000) == EOF) 
		st_p = close_stream(st_p);
	}

    /*	Now search for the "best" stream based on:			*/
    /*	    absolute start time (if any).				*/
    /*	    current input stream.					*/
    /*	    characteristics (if any) of current output stream.		*/
    /*	    specified bias.						*/
    /*	NOTE:  
    /*	    We should never be called with both a start time AND	*/
    /*	    a current input stream.					*/

    /*	If we have a current input stream, get the next block.		*/
    if (in_p != NULL) {
	/*  Get the next block of the current input stream.		*/
	if (position_stream (in_p, stime, etime, b2000_recnum) == EOF) {
	    /* EOF -- close it and unset current input stream.		*/
	    in_p = close_stream(in_p);
	    in_p = NULL;
	    /* Position all other input streams before selecting one.	*/
	    for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
		if (position_stream (st_p, stime, etime, b2000_recnum) == EOF) 
		    st_p = close_stream(st_p);
	    }
	}
    }

    /*	If we have a start time, look for the "best" stream that	*/
    /*  has the closest time >= start time.				*/
    
    best_diffnum = DIHUGE;
    if (stime != NULL) {
	for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
		if (position_stream (st_p, stime, etime, b2000_recnum) == EOF)  {
		/*  Possible candidate.					*/
		/*  Select "best" stream based on bias.			*/
		if (best_p == NULL) best_p = st_p;
		/* Is this stream better than the one that we have?	*/
		else switch (bias) {
		case LARGE_BLOCK_BIAS:
		    if (st_p->blksize > best_p->blksize) best_p = st_p;
		    break;
		case SMALL_BLOCK_BIAS:
		    if (st_p->blksize < best_p->blksize) best_p = st_p;
		    break;
		default:    /* Stay with what we have.			*/
		    break;
		}
	    }
	}
	/*  If we have chosen a "best" stream, return.			*/
	/*  Otherwise, ignore start time and continue the search.	*/
	if (best_p != NULL) return (best_p);
	else stime = NULL;
    }

    /*	If no start time and no current input stream, just search for	*/
    /*	for the earliest block.						*/
    if (stime == NULL && in_p == NULL) {
	for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
	    if (best_p == NULL) { best_p = st_p; continue; }
	    if ((tdif = tdiff(st_p->cur_hdr->begtime,best_p->cur_hdr->begtime)) < 0)
		best_p = st_p;
	}
	return (best_p);
    }

    /*	We have a current input stream and not explicit start time.	*/
    else {  /*	(stime == NULL && in_p != NULL)				*/
	/*  Blindly stay with the current input stream unless there is	*/
	/*  a non-acceptable gap in blockette record number.		*/
	if (in_p->cur_hdr->xm1 == NEXT2000(b2000_recnum)) {
	    return(in_p);
	}

	/*  The current data stream is NOT perfect.			*/
	/*  Position each other input stream.				*/
	/*  Search for the best one we can find.			*/
	best_p = in_p;
	best_diffnum = IHUGE;
	for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
	    if (position_stream (st_p, stime, etime, b2000_recnum) == EOF) {
		st_p = close_stream(st_p);
		continue;
	    }
	    /* Look for the data stream that has the closest next	*/
	    /* blockette record number.					*/
	    diffnum = DIFF2000(st_p->cur_hdr->xm1,b2000_recnum);
	    if (diffnum< best_diffnum) {
		best_p = st_p;
		best_diffnum = diffnum;
		continue;
	    }
	}
	return (best_p);
    }
}
