/************************************************************************/
/*  Command line parsing routines.					*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: parse_cmdline.c,v 1.2 2000/10/26 13:39:26 doug Exp $ ";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/stat.h>
#include <ctype.h>

#define BUF_SIZE 2048

#include "qlib2.h"

#include "params.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"

static int str_cp (char *dest, char **sp, char t);
static int num_cp (int *val, char **sp, int n);
static parse_filename (char *filename, char *pattern);

#define	argc	(*pargc)
#define	argv	(*pargv)

/************************************************************************/
/*  parse_cmdline:							*/
/*	Parse command line.						*/
/************************************************************************/
int parse_cmdline 
   (int		*pargc,		/* ptr to argc.				*/
    char	***pargv)	/* ptr to argv.				*/
{
    char    *span = NULL;
    int	    span_flag = 0;
    char    *p;
    INT_TIME	*pt;

    /* Variables needed for getopt. */
    extern char	*optarg;
    extern int	optind, opterr;
    int		c;

    int i;
    int	nfiles;
    char *blksize_str = NULL;
    char *strip_blockette_str = NULL;

    char *file_list_file =  NULL;
    struct stat stbuf;
    FILE *ListFILE;
    char line_buffer[BUF_SIZE];
    int s;
    int n;
    
    char *opts =  "hnTvcaeuHG:g:d:b:o:i:f:t:s:x:D:S:C:L:N:V:F:O:I:P:p:X:";

    cmdname = tail(*argv);
    seed_version = 2.3;
    default_network_id = DEFAULT_NETWORK;
    stream_tol = -1;
    block_tol = DEFAULT_BLOCK_TOL;
    output_blksize = DEFAULT_OUTPUT_BLKSIZE;
    output_data_wordorder = get_my_wordorder();
    input_data_fmt = STEIM1;
    duration_date_flag = 'n';		/* use nominal date in fname.	*/

    /*	Parse command line options.					*/
    while ( (c = getopt(argc,argv,opts)) != -1)
	switch (c) {
	case '?':
	case 'h':   print_syntax(cmdname,syntax,info); exit(0); break;
	case 'n':   ++no_output; break;
	case 'T':   ++exact_trim; break;
	case 'v':   ++verify_data; break;
	case 'c':   ++make_index; break;
	case 'a':   ++append_output; break;
	case 'e':   ++event_only_flag; break;
	case 'u':   ++unique_flag; break;
	case 'H':   ++print_file_hdr; break;
	case 'd':   debug_option = atoi(optarg); break;
/*::	case 'G':   stream_tol = atoi(optarg) * USECS_PER_TICK; break; */
	case 'g':   block_tol = atoi(optarg) * USECS_PER_TICK; break;
	case 'b':   blksize_str = optarg; break;
	case 'o':   outfile = optarg; break;
	case 'x':   strip_blockette_str = optarg; break;
	case 'S':   station_id = optarg; break;
	case 'C':   channel_id = optarg; break;
	case 'N':   network_id = optarg; break;
	case 'L':   location_id = optarg; break;
	case 'V':   seed_version = atof(optarg); break;
	case 'F':   file_list_file = optarg; break;
	case 'O':   output_data_wordorder = atoi(optarg); break;
	case 'I':   input_data_fmt = decode_data_format(optarg); break;
	case 'P':   file_duration = optarg; break;
	case 'p':   duration_date_flag = optarg[0]; break;
	case 'i':
	    /* Ignore some type of information. */
	    if (strcasecmp(optarg, "time") == 0) { ++ignore_hdr; break; }
	    fprintf (info, "Error: invalid ignore option: %s\n", optarg);
	    exit(1);
	    break;
	case 'f':
	    if ((pt = parse_date (optarg)) == NULL) {
		fprintf (info, "Error: invalid date: %s\n", optarg);
		exit(1);
	    }
	    start_time = *pt;
	    ++start_flag;
	    break;
	case 't':
	    if ((pt = parse_date (optarg)) == NULL) {
		fprintf (info, "Error: invalid date: %s\n", optarg);
		exit(1);
	    }
	    end_time = *pt;
	    ++end_flag;
	    break;
	case 's':
	    span = optarg; ++span_flag; break;
	case 'D':
	    /* Date display format.   */
	    if (strcmp (optarg, "j") == 0) date_fmt = JULIAN_FMT;
	    else if (strcmp (optarg, "j,") == 0) date_fmt = JULIAN_FMT_1;
	    else if (strcmp (optarg, "m") == 0) date_fmt = MONTH_FMT;
	    else if (strcmp (optarg, "m,") == 0) date_fmt = MONTH_FMT_1;
	    else {
		fprintf (info, "Error: invalid format for date display: %s\n", optarg);
		exit(1);
	    }
	    break;
	default:
	    fprintf (info, "Error: invalid option: -%c\n", c);
	    exit(1);
	}
    stripped_blockette = make_strip_list (strip_blockette_str, seed_version);
    if (make_index) ++no_output;
    if (file_duration) {
	if (parse_duration (file_duration) < 0) {
	    fprintf (info, "Error: invalid duration: %s\n", file_duration);
	    exit(1);
	}
	if (outfile) {
	    filename_pattern = outfile;
	    outfile = NULL;
	}
    }
    if (! (duration_date_flag == 'n' || duration_date_flag == 'd')) {
	fprintf (stderr, "Error: Invalid option: -p%c\n", duration_date_flag);
	exit(1);
    }
    if (no_output) outfile = "/dev/null";
    if (outfile == NULL && ! file_duration) info = stderr;
    if (end_flag && span_flag) {
	fprintf (stderr, "Error: span and endtime mutually exclusive\n");
	exit(1);
    }
    if (span_flag && !start_flag) {
	fprintf (stderr, "Error: span not valid without start time\n");
	exit(1);
    }
    if (span_flag && ! valid_span(span)) {
	    fprintf (stderr, "Error: invalid span specification: %s\n", span);
	    exit(1);
	}
    if (span_flag) {
	end_time = end_of_span (start_time, span);
	++end_flag;
    }
    if (blksize_str) {
	output_blksize = get_blksize(blksize_str);
	if (output_blksize <= 0 || output_blksize > MAX_BLKSIZE) {
	    fprintf (stderr, "Error: Invalid blksize: %s\n", blksize_str);
	    exit(1);
	}
	if (qverify) {
	    verify.blksize = output_blksize;
	    verify.blksize_flag++;
	}
    }
    if (input_data_fmt == UNKNOWN_DATATYPE) {
	fprintf (stderr, "Error: invalid input datatype specified\n");
	exit(1);
    }
    if (output_data_wordorder != SEED_BIG_ENDIAN && 
	output_data_wordorder != SEED_LITTLE_ENDIAN) {
	fprintf (stderr, "Error: Invalid wordorder: %d\n", output_data_wordorder);
	exit(1);
    }
    set_hdr_wordorder(output_data_wordorder);
    set_data_wordorder(output_data_wordorder);

    outmode = (append_output) ? "a" : "w";
    st_head_out.blksize = output_blksize;
    /* If file_duration, defer actual filename an opening until later.	*/
    if (file_duration) outfile = "";
    open_stream (&st_head_out, outfile, outmode);

    /*	Skip over all options and their arguments.			*/
    argv = &(argv[optind]);
    argc -= optind;

    /* If the user specified a file_list_file, then open the		*/
    /* file_list_file and read the names of the data files from it.	*/
    /* Otherwise, the remaining arguments are the input files.		*/
    /* If there is not file_list_file or data files on the cmdline,	*/
    /* then assume 1 file from stdin.					*/
    /* Open the files and allocate the appropriate data structures.	*/

    if (file_list_file) {
	if ((ListFILE = fopen(file_list_file, "r")) == NULL) {
	    fprintf (info, "Error: Unable to open file_list_file: %s\n", file_list_file);
	    exit(1);
	}
	while (fgets(line_buffer, BUF_SIZE - 1, ListFILE) != NULL) {
	    if (line_buffer[0] == '#')
		continue;  /* skip it, it's a comment */
	    /* remove the newline */
	    s = strlen(line_buffer);
	    line_buffer[s - 1] = '\0';
	    open_stream (&st_head_in, line_buffer, "r");
	}
	fclose(ListFILE);
    }
    else {
	nfiles = argc;
	do {
	    open_stream (&st_head_in, (nfiles > 0) ? *argv++ : NULL, "r");
	} while (--argc > 0);
    }

    return(0);
}

/************************************************************************/
/*  get_blksize:							*/
/*	Parse blksize_str and return blocksize.  Ensure it is valid, ie	*/
/*	it is a power of 2, and within the valid blksize boundaries.	*/
/*	Return 0 on error.						*/
/************************************************************************/
int get_blksize
   (char	*str)		/* string containing blksize.		*/
{
    int blksize, n, m;
    char *p;
    blksize = strtol (str, &p, 10);
    switch (*p) {
      case 0:	break;
      case 'k':
      case 'K':	blksize *= 1024; break;
      default:	blksize = 0;
    }
    if (blksize <= 0) return (0);
    n = (int)log2((double)blksize);
    m = (int)pow(2.0,(double)n);
    if (m != blksize) return (0);
    return (blksize);
}

/************************************************************************/
/*  parse_range:							*/
/*	Parse a range string of the format:				*/
/*	    i | i-j[,k-l]* | all					*/
/*	Return a ptr to the malloced range array.  Range array is	*/
/*	terminated by a (0,0) value.					*/
/************************************************************************/
RANGE *parse_range
   (char	*optarg)	/* string containing blockette range.	*/
{
    char *str = optarg;
    char *p;
    int first, last;
    int n = 0;
    RANGE *r = NULL;

    if (optarg == NULL) return (NULL);
    r = (RANGE *)malloc (sizeof(RANGE));
    if (strcasecmp(str, "all")==0) {
	r[n].first = 1;
	r[n].last = 65536;
	n++;
	r = (RANGE *)realloc (r, (n+1) * sizeof(RANGE));
    }
    else {
	while (*str != '\0') {
	    first = strtol (str, &p, 10);
	    if (first == 0) {
		fprintf (stderr, "Error: Invalid blockette list: %s\n", optarg);
		exit(1);
	    }
	    switch (*p) {
	      case ':':
	      case '-':
		str = ++p;
		last = strtol (str, &p, 10);
		break;
	      case ',':
	      case 0:
		last = first;
		break;
              default:
		fprintf (stderr, "Error: Invalid blockette list: %s\n", optarg);
		exit(1);
	    }
	    r[n].first = first;
	    r[n].last = last;
	    n++;
	    r = (RANGE *)realloc (r, (n+1) * sizeof(RANGE));
	    if (*p != '\0') ++p;
	    str = p;
	}
    }
    r[n].first = r[n].last = 0;
    return (r);
}

/************************************************************************/
/*  make_strip_list:							*/
/*	Create a list of blockettes to be stripped (removed).		*/
/************************************************************************/
RANGE *make_strip_list 
   (char	*str,		/* blockette list and/or range string.	*/
    double	version)	/* seed version number.			*/
{
    RANGE *r;
    int n, i;

    /* Parse string into an array of blockette ranges.			*/
    /* The array must terminate with a (0,0) range.			*/
    r = parse_range (str);
    if (r == NULL) {
	r = (RANGE *)malloc (sizeof(RANGE));
	r[0].first = r[0].last = 0;
    }

    /* Count the number of actual ranges from the string.		*/
    n = 0;
    for (n=0; r[n].first != 0; n++) ;

    /* Examine the pre-defined list of per-record blockettes.		*/
    /* If the seed version number of the blockette is greater than the	*/
    /* the desired seed version, add the blockette number onto the end	*/
    /* of the blockette range array.					*/
    for (i=0; per_record_blockette[i].blockette!=0; i++) {
	if (per_record_blockette[i].seed_version > version) {
	    r[n].first = r[n].last = per_record_blockette[i].blockette;
	    n++;
	    r = (RANGE *)realloc (r, (n+1) * sizeof(RANGE));
	}
    }
    /* Examine the pre-defined list of unique blockettes.		*/
    /* If the seed version number of the blockette is greager than the	*/
    /* the desired seed version, add the blockette number onto the end	*/
    /* of the blockette range array.					*/
    for (i=0; unique_blockette[i].blockette!=0; i++) {
	if (unique_blockette[i].seed_version > version) {
	    r[n].first = r[n].last = unique_blockette[i].blockette;
	    n++;
	    r = (RANGE *)realloc (r, (n+1) * sizeof(RANGE));
	}
    }

    /* Ensure that the last blockette range array entry is (0,0).	*/
    r[n].first = r[n].last = 0;
    return (r);
}

/************************************************************************/
/*  str_cp:								*/
/*	Copy a string until the terminator character or EOS is found.	*/
/*	Update src ptr.							*/
/*	Return:	0 on success, non-zero on error.			*/
/************************************************************************/
static int str_cp (char *dest, char **sp, char t) 
{
    while (**sp && **sp != t) {
	*dest++ = **sp;
	++(*sp);
    }
    *dest = '\0';
    return (0);
}

/************************************************************************/
/*  num_cp:								*/
/*	Copy a number for the specified number of digits.		*/
/*	Update src ptr.							*/
/*	Error if the specified number of digits are not found.		*/
/*	Return:	0 on success, non-zero on error.			*/
/************************************************************************/
static int num_cp (int *val, char **sp, int n) 
{
    static char str[BUF_SIZE];
    char *dest = str;
    while (n > 0 && isdigit(**sp)) {
	*dest++ = **sp;
	++(*sp);
	--n;
    }
    *dest = '\0';
    if (n > 0) return (-1);
    *val = atoi(str);
    return (0);
}

/************************************************************************/
/*  parse_filename:							*/
/*	Parse filename for verify version of the program.		*/
/*	Return:	0 on success, non-zero on failure.			*/
/************************************************************************/
static int parse_filename (char *filename, char *pattern)
{
    static char copy[BUF_SIZE];
    static char pat[BUF_SIZE];
    static char str[BUF_SIZE];
    char *p, *f, *s, *pyear_fn, *pdoy_fn;
    int n;

    strcpy (verify.filename, filename);
    strcpy (copy, filename);
    strcpy (pat, pattern);
    p = pat;

    /* Parse filename. */
    f = copy;
    /* Skip leading directory portion of filename. */
    if ((s=strrchr(f,'/'))) f = ++s;

    /* Using the pattern, parse the filename.	*/
    /*  NOTE: Allow %% to reduce to a single % 	*/
    /*  so that a pattern string can be printed	*/
    /*	in a printf string.			*/
    /*	%S or %s = station			*/
    /*	%N or %n = network			*/
    /*	%C or %c = channel			*/
    /*	%L or %l = location			*/
    /*	%Y	 = 4 digit year			*/
    /*	%j	 = 3 digit day of year		*/
    /*	%m	 = 2 digit month		*/
    /*	%d	 = 2 digit day of month		*/
    /*	%H	 = 2 digit hour			*/
    /*	%N	 = 2 digit minute		*/
    /*	%X or %x = match rest of token		*/

    /* Using pattern, pick up the specified components from filename.	*/
    /* Return an error if the filename does not match the pattern.	*/
    
    while (*p && *f) {
	switch (*p) {
	  case '%':
	    ++p;
	    if (*p == '%') ++p;		/* Treat %% as %.		*/
	    switch (*p) {
	      case 'S':
	      case 's':
		if (str_cp (str, &f, '.')) return (-1);
		if ((n=strlen(str)) > SDR_STATION_LEN) return (-1);
		if (! verify.station_flag) strcpy (verify.station_id, str);
		uppercase(verify.station_id);
		++verify.station_flag;
		++p;
		break;
	      case 'N':
	      case 'n':
		if (str_cp (str, &f, '.')) return (-1);
		if ((n=strlen(str)) > SDR_NETWORK_LEN) return (-1);
		if (! verify.network_flag) strcpy (verify.network_id, str);
		uppercase(verify.network_id);
		++verify.network_flag;
		++p;
		break;
	      case 'C':
	      case 'c':
		if (str_cp (str, &f, '.')) return (-1);
		if ((n=strlen(str)) > SDR_CHANNEL_LEN) return (-1);
		if (! verify.channel_flag) strcpy (verify.channel_id, str);
		uppercase(verify.channel_id);
		++verify.channel_flag;
		++p;
		break;
	      case 'L':
	      case 'l':
		if (str_cp (str, &f, '.')) return (-1);
		if ((n=strlen(str)) > SDR_LOCATION_LEN) return (-1);
		if (! verify.location_flag) strcpy (verify.location_id, str);
		uppercase(verify.location_id);
		++verify.location_flag;
		++p;
		break;
	      case 'Y':
		if (num_cp (&n, &f, 4)) return (-1);
		verify.year = n;
		++verify.year_flag;
		++p;
		break;
	      case 'J':
	      case 'j':
		if (num_cp (&n, &f, 3)) return (-1);
		if (n < 0 || n > 366) return (-1);
		verify.doy = n;
		++verify.doy_flag;
		++p;
		break;
	      case 'm':
		if (num_cp (&n, &f, 2)) return (-1);
		if (n < 0 || n > 12) return (-1);
		verify.month = n;
		++verify.month_flag;
		++p;
		break;
	      case 'd':
		if (num_cp (&n, &f, 2)) return (-1);
		if (n < 0 || n > 31) return (-1);
		verify.day = n;
		++verify.day_flag;
		++p;
		break;
	      case 'H':
		if (num_cp (&n, &f, 2)) return (-1);
		if (n < 0 || n > 23) return (-1);
		verify.hour = n;
		++verify.hour_flag;
		++p;
		break;
	      case 'M':
		if (num_cp (&n, &f, 2)) return (-1);
		if (n < 0 || n > 59) return (-1);
		verify.minute = n;
		++verify.minute_flag;
		++p;
		break;
	      case 'X':
	      case 'x':
		if (str_cp (str, &f, '.')) return (-1);
		++p;
	      default:
		return (-1);
	    }
	    break;
	  default:
	    if (*p != *f) return (-1);
	    ++p; ++f;
	    break;
	}
    }
    /* If we still have more pattern left, treat it as an error.    */
    if (*p) return (-1);
    return(0);
}
