/************************************************************************/
/*  Routine to unpack, trim and repack current input record.		*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: trim_data.c,v 1.2 2000/10/26 13:39:29 doug Exp $ ";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <math.h>

#include "qlib2.h"

#include "params.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"

/************************************************************************/
/*  trim_data:								*/
/*	Trim beginning and/or end of input data record(s).		*/
/*	Unpack all records, coalesce all blockettes, trim data,		*/
/*	optionally set new initial diff, and reblock into one or more	*/
/*	records.							*/
/*	It is assumed the ptr st_p->cur_hdr is same as hdrs[0],	so	*/
/*	st_p->cur_hdr is updated to point to the first new header.	*/
/************************************************************************/
void trim_data 
   (ST_INFO	*st_p,		/* ptr to input stream struct.		*/
    int		trim_through,	/* blockette recnum to trim through.	*/
    int		blksize,	/* desired output blksize.		*/
    RECORD_LIST	record_list[])	/* list of records.			*/
{
    int		    *p;
    int		    npts;		/* total # of data points.	*/
    int		    n;			/* index of next pt to output.	*/
    DATA_HDR	    *hdr;		/* current header.		*/
    DATA_HDR	    *master_hdr;	/* master composite hdr.	*/
    BS		    *bs;		/* ptr to blockette structure.	*/
    BS		    *next_bs;		/* ptr to next blockette struct.*/
    BS		    *delete_bs;		/* blockette structure to free.	*/
    BLOCKETTE_HDR   *bh;		/* ptr to blockette hdr.	*/
    int		    bl_type;		/* blockette type.		*/
    int		    bl_len;		/* blockette length.		*/
    int		    bl_space;		/* bl_len rounded to 4 bytes.	*/
    int		    bl_size;		/* total size of all blockettes.*/
    int		    bl_index;		/* index for blockette insert.	*/
    int		    h, i;		/* indices.			*/
    int		    hdr_size;		/* size of header + blockettes.	*/
    int		    max_data_bytes;	/* max # data bytes in record.	*/
    int		    data_offset;	/* offset (in bytes) of data.	*/
    int		    unique_info_added;	/* flag for unique info in frame*/
    int		    recnum;		/* b2000 record_num.		*/
    int		    xm1;		/* first b2000 record_num.	*/
    int		    xm2;		/* last b2000 record_num.	*/
    int		    ntrim;		/* number of b2000 to trim.	*/
    MS_ATTR	    attr;
    char	    debug_str[65536];

    /* Trimming or reblocking data may cause an expansion in the number	*/
    /* of bytes used to represent this input record.  The original	*/
    /* record may have already been unpacked and repacked into one or	*/
    /* more records.  We must therefore unpack all records on the list,	*/
    /* trim the data or fix the initial differences, and then repack	*/
    /* the data into one or more records.				*/
    /* Since each record needs to be constructed with the data located	*/
    /* at the appropriate offset within the IOB buffer, we allocate a	*/
    /* separate IOB for each new record that we recreate.		*/
    /*									*/
    /* It is the responsibility of the calling routine to free the 	*/
    /* DATA_HDRs and IOBs in the record_list.				*/

    /* Make a master hdr which is a copy of the first on the list.	*/
    /* This hdr will get a copy of every unique blockette from every	*/
    /* hdr on the list.							*/
    master_hdr = dup_data_hdr(record_list[0].hdr);
    delete_blockette (master_hdr, -1);
    master_hdr -> num_data_frames = 0;

    /* Loop over all records in the list, saving all unique blockettes	*/
    /* and unpacking all data frames.					*/
    /* Free the hdrs and iobs for the records after they are unpacked.	*/
    npts = 0;
    for (h=0; record_list[h].hdr!=NULL; h++) {
	merge_blockettes (master_hdr, record_list[h].hdr);
	st_p->cur_hdr = record_list[h].hdr;
	st_p->cur_hdr->num_samples = st_p->cur_hdr->num_data_frames = 0;
	st_p->cur_hdr->xn = st_p->cur_hdr->x0;
	free_data_hdr (record_list[h].hdr);
	free_iob (record_list[h].iob);
	record_list[h].hdr = NULL;
	record_list[h].iob = NULL;
    }
    add_required_blockettes(master_hdr);
    delete_unwanted_blockettes(master_hdr);

    /* Trim (ie delete) any 2000s if necessary.				*/
    ntrim = 0;
    if (trim_through != IGNORE2000) {
	for (bs = master_hdr->pblockettes; bs != NULL; bs = next_bs) {
	    bl_type = ((BLOCKETTE_HDR*)(bs->pb))->type;
	    next_bs = bs->next;
	    if (bl_type == 2000 && LE2000(((BLOCKETTE_2000 *)(bs->pb))->record_num,trim_through)) {
		if (! delete_pblockette (master_hdr, bs)) {
		    fprintf (info, "Error: unable to delete blockette 2000 in trim_data\n");
		    exit(1);
		}
		++ntrim;
	    }
	}
    }

    if (debug(DEBUG_STREAM)) {
	fprintf (info, "trimming input record: %d blockettes from head\n",
		 ntrim);
	fprintf (info, "trim, seq=%d, channel=%s, time=%s\n",
		 master_hdr->seq_no, master_hdr->channel_id, 
		 time_to_str(master_hdr->begtime,date_fmt));
	dump_blk(master_hdr, debug_str);
	fprintf (info, "master_hdr contains:\n%s",debug_str);
    }

    /* Repack the data into output frames into one or more records.	*/
    h = 0;
    n = 0;
    while ((bs=find_blockette(master_hdr, 2000)) != NULL) {
	unique_info_added = 0;
	
	/* Make a new header, and initialize it from the master hdr.	*/
	hdr = record_list[h].hdr = dup_data_hdr(master_hdr);
	delete_blockette (hdr, -1);
	record_list[h].iob = new_iob();
	hdr->num_samples = hdr->num_data_frames = hdr->num_blockettes = 0;
	hdr->first_data = hdr->first_blockette = 0;
	/* Explicitly set the desired output data_wordorder.		*/
	hdr->data_wordorder = output_data_wordorder;
	/* Explicitly set the desired blksize.				*/
	hdr->blksize = blksize;

	/* Copy and/or move all blockettes that fit from the master_hdr	*/
	/* over to this header.						*/
	/* Make 2 passes through the master blockette list --		*/
	/* first for blockettes that must be applied to every record,	*/
	/* then for unique blockettes assigned to only 1 record.	*/
	bl_size = 0;
	for (bs=master_hdr->pblockettes; bs!=NULL; bs=bs->next) {
	    bl_type = ((BLOCKETTE_HDR*)(bs->pb))->type;
	    bl_len = bs->len;
	    bl_space = bs->len + ((bs->len%4) ? 4-(bs->len%4) : 0);
	    if (! is_per_record_blockette(bl_type)) continue;
	    if (hdr->first_blockette == 0) hdr->first_blockette = 48;
	    if (bl_size + bl_space + hdr->first_blockette <= blksize) {
		bl_index = (bl_type == 1000) ? 0 : -1;
		add_blockette (hdr, bs->pb, bs->type, bs->len, bs->wordorder, bl_index);
		bl_size += bl_space;
	    }
	    else {
		fprintf (stderr, "Error: unable to fit minimal blockettes in output record\n");
		exit(1);
	    }
	}

	for (bs=master_hdr->pblockettes,delete_bs=NULL;
	     bs!=NULL; 
	     bs=bs->next,delete_pblockette(master_hdr,delete_bs),delete_bs=NULL) {
	    bl_type = ((BLOCKETTE_HDR*)(bs->pb))->type;
	    bl_len = bs->len;
	    bl_space = bs->len + ((bs->len%4) ? 4-(bs->len%4) : 0);
	    if (is_per_record_blockette(bl_type)) continue;
	    if (hdr->first_blockette == 0) hdr->first_blockette = 48;
	    if (bl_size + bl_space + hdr->first_blockette <= blksize) {
		bl_index = -1;
		if (bl_type == 2000) recnum = ((BLOCKETTE_2000 *)(bs->pb))->record_num;
		add_blockette (hdr, bs->pb, bs->type, bs->len, bs->wordorder, bl_index);
		delete_bs = bs;
		bl_size += bl_space;
		if (bl_type == 2000) {
		    if (unique_info_added == 0) xm1 = recnum;
		    xm2 = recnum;
		}
		++unique_info_added;
		if (debug(DEBUG_STREAM)) {
		    fprintf (info, "trim: seq=%d record=%d added blockette %d\n",
			     hdr->seq_no, h, recnum);
		    fflush (info);
		}
	    }
	    else {
		/* blockette 2000 cannot be taken out of order. */
		if (bl_type == 2000) break;
	    }
	}

	/* Ensure that we have room for something unique in this record-*/
	/* either unique blockette or data.  If not, the output		*/
	/* blocksize is too small, and we have deadlock.		*/
	attr = get_ms_attr (hdr);
	if (hdr->first_blockette == 0) hdr_size = 48;
	else hdr_size = hdr->first_blockette + bl_size;
	hdr_size = ((hdr_size+attr.alignment-1)/attr.alignment)*attr.alignment;
	data_offset = hdr_size;
	max_data_bytes = (blksize - hdr_size);
	if (unique_info_added == 0) {
	    fprintf (stderr, "Error: unable to fit any unique blockettes into output record\n");
	    exit(1);
	}

	/* We assume no data samples in these types of records.		*/
	hdr->num_samples = 0;
	hdr->num_data_frames = 0;
	hdr->first_data = 0;

	/*  Update the header information to reflect new data interval. */
	/*  We are now preserving for each repacked record:		*/
	/*	a.  num_ticks_correction.				*/
	/*	b.  The ACTIVITY_TIME_GAP, ACTIVITY_CALIB_PRESENT,	*/
	/*	    and ACTIVITY_EVENT_IN_PROGRESS flags.		*/
	/*	a.  hdrtime, which may or may not have			*/
	/*	    num_ticks_correction added in.			*/
	/*  Compute new begtime, hdrtime, and endtime in a manner that	*/
	/*  prevents overflows.						*/
	hdr->begtime = master_hdr->begtime;
	hdr->hdrtime = master_hdr->hdrtime;
	hdr->endtime = master_hdr->endtime;
	hdr->x0 = 0;
	hdr->xn = 0;
	hdr->xm1 = xm1;
	hdr->xm2 = xm2;
	hdr->num_ticks_correction = master_hdr->num_ticks_correction;
	if (h != 0) {
	    hdr->activity_flags &= (ACTIVITY_TIME_GAP | 
				    ACTIVITY_BEGINNING_OF_EVENT |
				    ACTIVITY_CALIB_PRESENT);
	}
	h++;
    }

    /* Set the stream's current header to point to the new header, and	*/
    /* free the master header that held the union of all blockettes.	*/
    st_p->cur_hdr = record_list[0].hdr;
    st_p->iob = record_list[0].iob;
    free_data_hdr(master_hdr);

    if (debug(DEBUG_STREAM)) {
	fprintf (info, "trim: data repacked into %d records\n", h);
    }
    return;
}
